/*
-- Time-stamp: <Mon Dec 12 2005 20:21:40 hwloidl>
-- $Id: PPIntm.java,v 1.3 2005/12/08 23:23:26 hwloidl Exp $
--
-- A Compiler for MiniJava: 
--  Pretty printer for intermediate code
-- Practical on compiler design, LMU 2005/6
-- Based on Appels book on "Modern Compiler Implementation"
-- ---------------------------------------------------------------------------
*/

class PPIntm implements IntmVisitor<String> {
    // -----------------------------------------------------------------------
    // visitor for pretty printing the intermediate code

    public String visit(CONST c)  {
        return c.value.toString();
    }
    public String visit (NAME c) {
        return c.label.toString();
    }
    public String visit (TEMP c) {
        return c.temp.toString();
    }
    public String visit (BINOP c) {
      String lhs = c.left.accept(this);
      String rhs = c.right.accept(this);
      String op;
      switch (c.op) {
       case BINOP.PLUS:  if (c.right instanceof CONST && ((CONST)c.right).value<0) {  
                            op = "-"; 
                            rhs = (new CONST((-1)*((CONST)c.right).value)).accept(this); 
                         } else { 
                            op="+"; 
                         };  
                         break;
       case BINOP.MINUS: if (c.right instanceof CONST && ((CONST)c.right).value<0) {  
                           op = "+"; 
                           rhs = (new CONST((-1)*((CONST)c.right).value)).accept(this); 
                         } else { 
                           op="-"; 
                         };  
                         break;
       case BINOP.MUL: op = "*";  break;
       case BINOP.DIV: op = "/";  break;
       case BINOP.AND: op = "&&";  break;
       case BINOP.OR: op = "||";  break;
       case BINOP.LSHIFT: op = "<<";  break;
       case BINOP.RSHIFT: op = ">>";  break;
       case BINOP.ARSHIFT: op = ">|";  break;
       case BINOP.XOR: op = "^^";  break;
       default: op = "??";
      }
      return "("+lhs+op+rhs+")";
    }

  public String visit(MEM  c){
    return "MEM("+c.addr.accept(this)+")";
  }
  public String visit(CALL c){
    return "CALL "+"L"+c.func.accept(this)+"("+c.args.accept(new PPIntmExpList())+")";
  }
  public String visit(ESEQ c){
    String s = "{ "+c.stm.accept(this)+","+c.exp.accept(this)+" }";
    return s;
  }
  public String visit(MOVE c) {
    return c.dest.accept(this)+" <- "+c.src.accept(this);
  }
  public String visit(EXP  c){
    return c.exp.accept(this);
  }
  public String visit(JUMP c){
    return "JUMP "+c.dest.accept(this)+" ["+c.dest.accept(this)+"]";
  }
  public String visit(CJUMP c){
    String lhs = c.left.accept(this);
    String rhs = c.right.accept(this);
    String op;
    switch (c.rel) {
    case CJUMP.NE: op = "!="; break;
    case CJUMP.LT: op = "<"; break;
    default: op= "??";
    }
    return "CJUMP (" + lhs + op + rhs +") ? "+c.trueLab.toString()+" : "+c.falseLab.toString();
  }
  public String visit(SEQ  c){
    
    return c.first.accept(this)+"\n"+c.second.accept(this);
  }
  public String visit(LABEL c){
    return c.label.toString()+":";
  }

  public String visit(Exp c) {
    if (c instanceof CONST) {
      return visit((CONST)c);
    } else if (c instanceof NAME) {
      return visit((NAME)c);
    } else if (c instanceof TEMP) {
      return visit((TEMP)c);
    } else if  (c instanceof BINOP) {
      return visit((BINOP)c);
    } else if (c instanceof MEM) {
      return visit((MEM)c);
    } else if (c instanceof CALL) {
      return visit((CALL)c);
    } else if (c instanceof ESEQ) {
      return visit((ESEQ)c);
    } else {
      System.out.println("PANIC: illegal instance for "+c.toString());
    }
    return "ERROR: illegal instance for"+c.toString();
  }
  public String visit(Stm c) {
    if (c instanceof MOVE) {
      return visit((MOVE)c);
    } if (c instanceof EXP) {
      return visit((EXP)c);
    } else if (c instanceof JUMP) {
      return visit((JUMP)c);
    } else if (c instanceof CJUMP) {
      return visit((CJUMP)c);
    } else if (c instanceof SEQ) {
      return visit ((SEQ)c);
    } else if (c instanceof LABEL) {
      return visit ((LABEL)c);
    } else {
      System.out.println("PANIC: illegal instance for"+c.toString());
    }
    return "ERROR: illegal instance for "+c.toString();
  }
}

class PPIntmExpList implements ListVisitor<String,Exp> {
  public String visit(NIL<Exp> x) {
    return "";
  }
  public String visit(CONS<Exp> x) {
    if (x.tl instanceof NIL) {
      return x.hd.accept(new PPIntm());
    } else {      
      return x.hd.accept(new PPIntm())+", "+x.tl.accept(this);
    }
  }
}
